/****************************************************************************
*                      RECAFOH (pronounced REC-ah-foe)                      *
*                                                                           *
*           Written By Benjamin Lee Wiebe <itsme@benjiwiebe.com>            *
*                                                                           *
*          Remotely Execute a Command in an Alias File On the Host          *
****************************************************************************/
#ifdef HAVE_CONFIG_H
#include "fix_configs.h"
#endif
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <netdb.h>
#include <stdio.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/wait.h>
#include <assert.h>
#include "message.h"
#include "service.h"

// start - wrapper functions
int xunlink(const char*);
// end - wrapper functions
void ring_all_users(void);
void print_help(void);

void mem_free_all(void);
void sig_handler(int);
void sig_chld(int);

char *prg_invocname = NULL;

int port = 2012;
#if defined(USE_SYSTEMD)
const char * const service_filename = {"/lib/systemd/system/recafohd.service"};
#elif defined(USE_SYSVINIT)
const char * const service_filename = {"/etc/init.d/recafoh"};
#else
const char * const service_filename = {""};
#endif
#if defined(USE_SYSTEMD) || defined(USE_SYSVINIT)
const char * const shortopts = {"fnkp:irvwesh"};
#else
const char * const shortopts = {"fnkp:vwesh"};
#endif
const char *pid_filename = {"/var/run/recafohd.pid"};
const char *conf_filename = {"/etc/recafoh/recafohd.conf"};

int main(int argc, char *argv[])
{
    static struct option long_options[] =
    {
        {"forking", no_argument, 0, 'f'},
        {"noforking", no_argument, 0, 'n'},
        {"kill-existing", no_argument, 0, 'k'},
        {"port", required_argument, 0, 'p'},
        #if defined(USE_SYSTEMD) || defined(USE_SYSVINIT)
        {"service-install", no_argument, 0, 'i'},
        {"service-remove", no_argument, 0, 'r'},
        #endif
        {"verbose", no_argument, 0, 'v'},
        {"warnings", no_argument, 0, 'w'},
        {"errors", no_argument, 0, 'e'},
        {"silent", no_argument, 0, 's'},
        {"help", no_argument, 0, 'h'}
    };

    int option = 0;

    int itmp = 0; /* variables with no dedicated purpose */
    char *cptmp = NULL;

    bool f_fork = false, f_inst = false, f_rm = false; /* flags set in switch(option): fork, service install, service remove */

    atexit(mem_free_all);
    signal(SIGTERM, sig_handler);
    signal(SIGHUP, sig_handler);

    /* Set prg_invocname for functions that need to know the contents of argv[0] */
    prg_invocname = (char*)malloc(strlen(argv[0])+1);
    memset(prg_invocname, 0, BUFSIZ);
    strcpy(prg_invocname, argv[0]);

    while((option = getopt_long_only(argc, argv, shortopts, long_options, 0)) != (-1))
    {
        switch(option)
        {
        case 'f':
            f_fork = true;
            break;
        case 'n':
            f_fork = false;
            break;
        case 'k':
            m_ferror(EXIT_FAILURE, "--kill-existing/-k not yet implemented.");
            break;
        case 'p':
            assert(optarg != (char*)0);
            port = atoi(optarg);
            break;
	#if defined(USE_SYSTEMD) || defined(USE_SYSVINIT)
        case 'i':
            f_inst = true;
            if(f_rm)
                m_ferror(EXIT_FAILURE, "%s: --service-remove and --service-install can not both be specified.", prg_invocname);
            break;
        case 'r':
            f_rm = true;
            if(f_inst)
                m_ferror(EXIT_FAILURE, "%s: --service-remove and --service-install can not both be specified.", prg_invocname);
            break;
	#endif
        case 'v':
            m_set_verbosity(V_VERBOSE);
            break;
        case 'w':
            m_set_verbosity(V_WARNINGS);
            break;
        case 'e':
            m_set_verbosity(V_ERRORS);
            break;
        case 's':
            m_set_verbosity(V_SILENT);
            break;
        case 'h':
            print_help();
            exit(EXIT_SUCCESS);
        default:
            exit(EXIT_FAILURE);
        }
    }
    if(f_fork)
    {
        FILE *pidfile = NULL;
        pidfile = fopen(pid_filename, "r");
        if(pidfile != NULL)
        {
            cptmp = (char*)malloc(15);
            fscanf(pidfile, "%s", cptmp); /* read in the PID */
            itmp = atoi(cptmp); /* convert it to an int */
            free(cptmp); /* free cptmp; we have the PID in itmp */
            cptmp = (char*)malloc(30);
            snprintf(cptmp, 29, "/proc/%d/environ", itmp); /* make cptmp = path to programs environ, just to see if program is running */
            FILE *procpid = fopen(cptmp, "r");
            if(procpid == NULL)
            {
                m_warning("%s: PID file still exists and process is not running. Deleting old PID file.\r\n", argv[0]);
                if(xunlink(pid_filename) < 0)
                {
                    m_error("%s: Failed to delete PID file '%s': %s.\r\n", argv[0], pid_filename, strerror(errno));
                    if((errno == EACCES)&&(geteuid()!=0)) /* error is access denied, and we are not root */
                    {
                        m_warning("%s: You really should be root to run this program as a daemon.\r\n", argv[0]);
                    }
                };
            }
            else
                m_warning("Another server is running. PID = %d.\r\n", itmp);
        }
        int pid = fork();
        int save_err = 0;
        if(pid != 0)
        {
            m_info("RECAFOH daemon started; PID = %d.\r\n", pid);
            pidfile = fopen(pid_filename, "w+");
            if(!pidfile)
            {
                save_err = errno;
                m_warning("%s: Warning - Could not open pid file '%s': %s.\r\n", argv[0], pid_filename, strerror(errno));
                if((save_err==EACCES)&&(geteuid()!=0))
                {
                    m_warning("%s: You really should be root to run this program as a daemon.\r\n", argv[0]);
                }
            }
            else
            {
                fprintf(pidfile, "%d\r\n", pid);
                fclose(pidfile);
            }
            exit(EXIT_SUCCESS);
        }
    }
    else if(f_inst)
    {
        install_service(argv[0], service_filename); /* don't use preprocessor statements to check whether systemd or sysvinit or nothing is used, this is handled by systemd.c/sysv-init.c */
        exit(EXIT_SUCCESS);
    }
    else if(f_rm)
    {
        remove_service(argv[0], service_filename); /* don't use preprocessor statements to check whether systemd or sysvinit or nothing is used, this 
is handled by systemd.c/sysv-init.c */
        exit(EXIT_SUCCESS);
    }

    signal(SIGCHLD, sig_chld);

    waitpid(-1, NULL, WNOHANG); // zombie catcher

    int sock;
    struct sockaddr_in meserv;
    socklen_t meserv_length;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0)
    {
        m_ferror(EXIT_FAILURE, "%s: Could not open socket on port %d: %s.\r\n", argv[0], port, strerror(errno));
    }

    meserv_length = sizeof(meserv);
    memset(&meserv, 0, meserv_length);
    meserv.sin_family=AF_INET;
    meserv.sin_port=htons(port);
    inet_pton(AF_INET, "0.0.0.0", &meserv.sin_addr);

    if(bind(sock, (struct sockaddr*)&meserv, meserv_length) < 0)
    {
        m_ferror(EXIT_FAILURE, "%s: Could not bind to port %d: %s.\r\n", argv[0], port, strerror(errno));
    }

    if(listen(sock, 1) < 0) /* have a max of queued connections */
    {
        m_ferror(EXIT_FAILURE, "%s: Could not listen on port %d: %s.\r\n", argv[0], port, strerror(errno));
    }
    
    int sockfd = -1, n = 0, buflen;
    char *buf = NULL;
    while(1)
    {
        sockfd = accept(sock, (struct sockaddr*)&meserv, &meserv_length);
        if(sockfd < 0)
        {
            m_ferror(EXIT_FAILURE, "%s: Could not accept connection on port %d: %s.\r\n", argv[0], port, strerror(errno));
        }
        buf = (char*)malloc(256);
        memset(buf, 0, 256);
        n = read(sockfd, buf, 255);
        if(n < 0)
        {
            m_ferror(EXIT_FAILURE, "%s: Read failed on port %d: %s.\r\n", argv[0], port, strerror(errno));
        }

        /* remove end-of-line character(s) */
        buflen = strlen(buf) - 1;
        if(buf[buflen] == '\n')
            buf[buflen] = '\0';
        buflen = strlen(buf) -1;
        if(buf[buflen] == '\r')
            buf[buflen] = '\0';

        if(strcmp("ring_users", buf))
        {
            cptmp = (char*)malloc(BUFSIZ);
            memset(cptmp, 0, BUFSIZ);
            snprintf(cptmp, BUFSIZ, "Not a valid command: '%s'.\r\n", buf);
            n = write(sockfd, cptmp, strlen(cptmp));
            if(n < 0)
            {
                m_ferror(EXIT_FAILURE, "%s: Read failed on port %d: %s.\r\n", argv[0], port, strerror(errno));
            }
            free(cptmp);
            cptmp = NULL;
        }
        else
        {
            ring_all_users();
            n = write(sockfd, "Ringing users... Done.\n", 23);
            if(n < 0)
            {
                m_ferror(EXIT_FAILURE, "%s: Read failed on port %d: %s.\r\n", argv[0], port, strerror(errno));
            }
        }
        free(buf);
        close(sockfd);   
    }
    return 0;
}

void ring_all_users(void)
{
    int ret = fork();
    if(ret == -1)
        m_error("fork: Could not fork process: %s.\r\n", strerror(errno));
    else if(ret == 0)
        execlp("wall", "ring_users", NULL);
    else
        m_info("'wall' invoked: PID=%d.\r\n", ret);
}

void print_help(void)
{
    printf("Usage: %s [options]\r\n", prg_invocname);
    printf("  -f, --forking\t\tStart another process with fork() and exit (daemonize)\r\n");
    printf("  -n, --noforking\tThe opposite of --forking; run in foreground\r\n");
    printf("  -k, --kill-existing\tKill existing RECAFOH server (uses recafohd.pid) (Not implemented)\r\n");
    printf("  -p, --port <port>\tThe port number to listen on\r\n");
    printf("  -c, --command <cmd>\tUse <cmd> as the command to send\r\n");
    #if defined(USE_SYSTEMD) || defined(USE_SYSVINIT)
    printf("  -i, --service-install\tCreate a file to make recafohd start on boot (init script or systemd .service file)\r\n");
    printf("  -r, --service-remove\tRemoves the file created by --service-install\r\n");
    #endif
    printf("  -v, --verbose\t\tPrint everything\r\n");
    printf("  -w, --warnings\tPrint warnings and errors (default)\r\n");
    printf("  -e, --errors\t\tPrint the errors and nothing else\r\n");
    printf("  -s, --silent\t\tPrint nothing at all\r\n");
    printf("  -h, --help\t\tPrint this help message and exit\r\n");
}

void mem_free_all(void)
{
    /*FIXME Why is this causing a segfault?
    if(prg_invocname != NULL)
        free(prg_invocname);*/
}

void sig_handler(int signum)
{
    if(signum == SIGTERM)
    {
        m_info("%s: Caught SIGTERM, termination signal; Exiting...\r\n", prg_invocname);
        if(xunlink("/var/run/recafohd.pid") < 0)
        {
            m_error("%s: Could not delete PID file '/var/run/recafohd.pid': %s.\r\n", prg_invocname, strerror(errno));
            if((errno == EACCES) && (geteuid() != 0))
            {
                m_warning("%s: You really should have been root to run this program as a daemon, but too late now. ;)\r\n", prg_invocname);
            }
        }
    }
    else if(signum == SIGHUP)
    {
        m_info("%s: Caught SIGHUP, hang-up signal; Exiting...\r\n", prg_invocname);
        if(xunlink("/var/run/recafohd.pid") < 0)
        {
            m_error("%s: Could not delete PID file '/var/run/recafohd.pid': %s.\r\n", prg_invocname, strerror(errno));
            if((errno == EACCES) && (geteuid() != 0))
            {
                m_warning("%s: You really should have been root to run this program as a daemon, but too late now. ;)\r\n", prg_invocname); /* TODO Add easter egg here: print :), backspace, print ;), backspace, print :) (animated wink) */
            }
        }
    }
    exit(EXIT_FAILURE);
}

void sig_chld(int sig_no)
{
    signal(SIGCHLD , sig_chld);
    int stat = sig_no; // meaningless - suppress 'unused' errors
    wait(&stat);
}
